"""Handle an outbound All-Link Broadcast message to a group."""

from abc import ABCMeta

from ...address import Address
from ...constants import MessageFlagType
from .. import ack_handler
from ..outbound_base import OutboundHandlerBase


class AllLinkBroadcastCommandHandlerBase(OutboundHandlerBase):
    """Abstract base class for outbound All-Link broadcast message handling."""

    __meta__ = ABCMeta

    def __init__(self, topic, group):
        """Init the DirectCommandHandlerBase class."""
        self._address = Address(bytearray([0x00, 0x00, group]))
        super().__init__(
            topic=topic,
            address=self._address,
            group=None,
            message_type=MessageFlagType.ALL_LINK_BROADCAST,
        )

    async def async_send(self, **kwargs):
        """Send the command and wait for a direct_nak."""
        return await super().async_send(address=self._address, **kwargs)

    # pylint: disable=arguments-differ
    @ack_handler
    def handle_ack(self, cmd1, cmd2, user_data):
        """Handle the message ACK."""
        super().handle_ack(cmd1=cmd1, cmd2=cmd2, user_data=user_data)



"""Send an All-Link Cleanup ACK message to a device."""

from ...constants import MessageFlagType
from .direct_command import DirectCommandHandlerBase


class AllLinkCleanupAckCommandHandlerBase(DirectCommandHandlerBase):
    """Abstract base class for outbound All-Link Cleanup ACK message handling."""

    def __init__(self, topic, address, group):
        """Init the DirectCommandHandlerBase class."""
        super().__init__(
            topic=topic,
            address=address,
            group=group,
            message_type=MessageFlagType.ALL_LINK_CLEANUP_ACK,
        )

    async def async_send(self, **kwargs):
        """Send the command and wait for a direct_nak."""
        return await super().async_send(**kwargs)



"""Insteon message and command handlers."""
import logging
from functools import partial, wraps

from ..address import Address
from ..constants import MessageFlagType
from ..utils import build_topic, subscribe_topic

TIMEOUT = 3  # Time out between ACK and Direct ACK
_LOGGER = logging.getLogger(__name__)


def _register_handler(
    func,
    presets=None,
    topic=None,
    prefix=None,
    address=None,
    group=None,
    message_type=None,
):
    """Register the function with a topic."""
    presets = presets or {}
    topic = presets.get("topic") or topic
    address = presets.get("address") or address
    get_group = presets.get("group", -1)
    group = get_group if get_group != -1 else group
    message_type = presets.get("message_type") or message_type
    full_topic = build_topic(
        topic=topic,
        prefix=presets.get("prefix"),
        address=address,
        group=group,
        message_type=message_type,
    )
    subscribe_topic(func, full_topic)


def _setup_handler(reg_func, func):
    """Set up the handler function."""

    @wraps(func)
    def _wrapper(self, *args, **kwargs):
        """Wrap a function."""
        return func(self, *args, **kwargs)

    _wrapper.register_handler = reg_func
    return _wrapper


def inbound_handler(func):
    """Register any inbound message handler."""
    reg_func = partial(
        _register_handler,
    )
    return _setup_handler(reg_func, func)


def ack_handler(func):
    """Register the message ACK handler."""
    reg_func = partial(_register_handler, presets={"prefix": "ack"})
    return _setup_handler(reg_func, func)


def nak_handler(func):
    """Register the message NAK handler."""
    reg_func = partial(_register_handler, presets={"prefix": "nak"})
    return _setup_handler(reg_func, func)


def direct_ack_handler(func):
    """Register the DIRECT_ACK response handler."""
    reg_func = partial(
        _register_handler,
        presets={"group": None, "message_type": MessageFlagType.DIRECT_ACK},
    )
    return _setup_handler(reg_func, func)


def status_handler(func):
    """Register the status response handler."""

    def register_status(instance_func, address):
        # This registers all messages for a device but only triggers on
        # status messages if they return within the TIMEOUT period
        address = Address(address)
        subscribe_topic(instance_func, address.id)

    func.register_status = register_status
    return func


def direct_nak_handler(func):
    """Register the DIRECT_NAK response handler."""
    reg_func = partial(
        _register_handler,
        presets={"group": None, "message_type": MessageFlagType.DIRECT_NAK},
    )
    return _setup_handler(reg_func, func)


def broadcast_handler(func):
    """Register the BROADCAST message handler."""

    def _register_broadcast_handler(
        func, topic, address=None, group=None, message_type=None
    ):
        """Register the function with a topic."""
        broadcast_topic = build_topic(
            topic=topic,
            prefix=None,
            address=address,
            group=group,
            message_type=MessageFlagType.BROADCAST,
        )
        all_link_topic = build_topic(
            topic=topic,
            prefix=None,
            address=address,
            group=group,
            message_type=MessageFlagType.ALL_LINK_BROADCAST,
        )
        subscribe_topic(func, broadcast_topic)
        subscribe_topic(func, all_link_topic)

    reg_func = partial(_register_broadcast_handler)
    return _setup_handler(reg_func, func)


def all_link_cleanup_ack_handler(func):
    """Register the all_link_cleanup ACK response handler."""
    reg_func = partial(
        _register_handler,
        presets={"message_type": MessageFlagType.ALL_LINK_CLEANUP_ACK},
    )
    return _setup_handler(reg_func, func)


def all_link_cleanup_nak_handler(func):
    """Register the all_link_cleanup NAK response handler."""
    reg_func = partial(
        _register_handler,
        presets={"message_type": MessageFlagType.ALL_LINK_CLEANUP_ACK},
    )
    return _setup_handler(reg_func, func)



"""Manage outbound Png command to a device."""
from ...topics import PING
from .. import direct_ack_handler
from .direct_command import DirectCommandHandlerBase


class PingCommand(DirectCommandHandlerBase):
    """Manage an outbound Ping command to a device."""

    def __init__(self, address):
        """Init the PingCommand class."""
        super().__init__(topic=PING, address=address)

    @direct_ack_handler
    def handle_direct_ack(self, cmd1, cmd2, target, user_data, hops_left):
        """Handle the ping response direct ACK."""
        self._call_subscribers()
        super().handle_direct_ack(cmd1, cmd2, target, user_data, hops_left)



"""Manage outbound ON command to a device."""
from pyinsteon.constants import MessageFlagType, ResponseStatus

from ... import pub
from ...topics import STATUS_REQUEST
from .. import ack_handler, status_handler
from .direct_command import DirectCommandHandlerBase


class StatusRequestCommand(DirectCommandHandlerBase):
    """Manage an outbound Status command to a device.

    TODO Confirm that the status command is best with a single status handler
    rather than a handler per status command (ie. one for state 1 and one for
    state 2)
    """

    def __init__(self, address, status_type: int = 0):
        """Init the OnLevelCommand class."""
        super().__init__(topic=STATUS_REQUEST, address=address)
        self._status_active = False
        self._status_type = status_type
        if status_type:
            self._subscriber_topic = f"{self._subscriber_topic}_{status_type}"

    @property
    def status_type(self):
        """Return the type of status message."""
        try:
            return self._status_type
        except AttributeError:
            return 0

    @property
    def status_active(self):
        """Return if the status command is active."""
        try:
            return self._status_active
        except AttributeError:
            return False

    @status_active.setter
    def status_active(self, value: bool):
        """Set if the status command is active."""
        self._status_active = bool(value)

    # pylint: disable=arguments-differ, useless-super-delegation
    async def async_send(self):
        """Send the ON command async."""
        response = await super().async_send(status_type=self._status_type)
        # return response
        self._status_active = False
        return response

    @ack_handler
    def handle_ack(self, cmd1, cmd2, user_data):
        """Handle the message ACK."""
        if cmd2 == self.status_type:
            self.status_active = True
            super().handle_ack(cmd1, cmd2, user_data)

    @status_handler
    def handle_direct_ack(self, topic=pub.AUTO_TOPIC, **kwargs):
        """Handle the ON response direct ACK.

        This handler listens to all topics for a device therefore we need to
        confirm the message is a status response.
        """
        if not self.status_active:
            return

        msg_type = topic.name.split(".")[-1]
        if msg_type != str(MessageFlagType.DIRECT_ACK):
            return

        self._message_response.put_nowait(ResponseStatus.SUCCESS)

        cmd1 = kwargs.get("cmd1")
        cmd2 = kwargs.get("cmd2")
        if cmd2 is not None:
            self._call_subscribers(db_version=cmd1, status=cmd2)



"""Command line tools to interact with the Insteon devices."""

from .. import devices
from ..constants import RAMP_RATES, ALDBStatus, DeviceCategory
from ..managers.scene_manager import async_add_device_to_scene
from ..utils import seconds_to_ramp_rate
from .advanced import AdvancedTools
from .tools_base import ToolsBase


class ToolsAldb(ToolsBase):
    """Command class to test interactivity."""

    async def do_load_aldb(self, *args, **kwargs):
        """Load the All-Link Database of a device.

        Usage:
            load_aldb <ADDRESS>|all y|n  Load one or all devices (can be the modem address)

        To clear the current ALDB and reload from the device, enter `y` as the second argment.
        Otherwise enter `n`.
        """
        args = args[0].split()
        try:
            address = args[0]
        except IndexError:
            address = None

        try:
            refresh_yn = args[1]
            refresh = refresh_yn.lower() == "y"
        except IndexError:
            refresh_yn = ""

        addresses = await self._get_addresses(
            address=address, allow_cancel=True, allow_all=True, match_device=True
        )
        if not addresses:
            return

        if devices[addresses[0]] != devices.modem or len(addresses) > 1:
            if not refresh_yn:
                refresh_yn = await self._get_char(
                    "Clear existing records and reload (y/n)",
                    default="n",
                    values=["y", "n"],
                )
        refresh = refresh_yn.lower() == "y"

        battery_devices = []
        for address in addresses:
            if devices[address].is_battery:
                battery_devices.append(address)
            # Only load the modem ALDB if explicitly asked
            if devices[address] == devices.modem and len(addresses) == 1:
                await devices.modem.aldb.async_load()
            elif devices[address].cat == 0x03:
                pass
            else:
                # tasks.append(devices[address].aldb.async_load(refresh=refresh))
                await devices[address].aldb.async_load(refresh=refresh)

        if battery_devices:
            self._log_stdout("The following devices are battery operated.")
            self._log_stdout("They will load in the background when they wake up.")
        for address in battery_devices:
            self._log_stdout(f"    - {address}")

        # if the device did not load the first time, try one more time with refresh
        for address in addresses:
            if (
                devices[address] != devices.modem
                and devices[address].aldb.status != ALDBStatus.LOADED
                and not devices[address].is_battery
            ):
                await devices[address].aldb.async_load(refresh=refresh)

    async def do_print_aldb(self, *args, **kwargs):
        """Print the records in an All-Link Database.

        Usage:
            print_aldb <ADDRESS>|all
        """
        await self._print_aldb(*args)

    async def do_add_default_links(self, *args, **kwargs):
        """Add default links between a device and the modem.

        Usage:
            add_default_links <ADDRESS>
        """

        args = args[0].split()
        try:
            address = args[0]
        except IndexError:
            address = None

        addresses = await self._get_addresses(
            address=address, allow_all=False, allow_cancel=True, match_device=True
        )
        if not addresses:
            return
        device = devices[addresses[0]]
        self._log_command(f"add_default_links {addresses[0]}")
        await device.async_add_default_links()

    async def do_add_device_to_scene(self, *args, **kwargs):
        """Add a device to a scene.

        Usage:
            add_device_to_scene <ADDRESS> <SCENE NUMBER> [<ON LEVEL>] [<RAMP RATE>] [<BUTTON>] | [Data1] [Data2] [Data3]
        <ADDRESS>: The device address such as 1a.2b.3c
        <SCENE NUMBER>: Value from 25 to 255.

        For Device type 1:
        <ON LEVEL>: (Optional) Value from 0 (off) - 255 (full on).
                    For dimmable devices any number from 0 to 255 is allowable.
                    Default is 255.
        <RAMP RATE>: 0.1 seconds to 480 seconds (8 minutes)
                     Default is 0.5 seconds
        <BUTTON>: (Optional) The button or group number of the device to change as part of the scene.
                    Valid values are device dependant.
                    Default is 1.

        for Device type 2:
        <Data1>: (Optional) Value from 0 (off) - 255 (full on).
                    For on/off devices only 0 and 255 are allowed.
                    Default is 255.
        <Data2>: Data field 2. Default is 0. Typically, this is not used by device type 2.
        <BUTTON>: (Optional) The button or group number of the device to change as part of the scene.
                    Valid values are device dependant.
                    Default is 1.

        For all other device types:
        <Data1>: Data field 1. Any value from 0 - 255 are allowed. Default is 255.
        <Data2>: Data field 2 Any value from 0 - 255 are allowed. Default is 0.
        <Data3>: Data field 3  Any value from 0 - 255 are allowed. Default is 1.

        KeyPadLinc devices will not respond correctly to scenes in this way other than the main power.
        """

        args = args[0].split()
        try:
            address = args[0]
        except IndexError:
            address = None

        try:
            scene = int(args[1])
        except (IndexError, ValueError):
            scene = None

        try:
            data1 = int(args[2])
        except (IndexError, ValueError):
            data1 = None

        try:
            data2 = int(args[3])
        except (IndexError, ValueError):
            data2 = None

        try:
            data3 = int(args[4])
        except (IndexError, ValueError):
            data3 = None

        addresses = await self._get_addresses(
            address=address, allow_all=False, allow_cancel=True, match_device=True
        )
        if not addresses:
            return

        device = devices[addresses[0]]

        if not scene:
            scene = await self._get_int(
                "Scene number or blank to cancel",
                values=range(25, 256),
            )
            if not scene:
                return

        if data1 is None:
            if device.cat == DeviceCategory.DIMMABLE_LIGHTING_CONTROL:
                data1 = await self._get_int(
                    "On level", default=255, values=range(0, 256)
                )
            elif device.cat == DeviceCategory.SWITCHED_LIGHTING_CONTROL:
                data1 = await self._get_int("On level", default=255, values=[0, 255])
            else:
                data1 = await self._get_int("Data1", default=255, values=range(0, 255))

        if device.cat == DeviceCategory.DIMMABLE_LIGHTING_CONTROL:
            if data2 is None:
                try:
                    data2_seconds = float(args[3])
                except (IndexError, ValueError):
                    data2_seconds = None
                if data2_seconds is None:
                    data2_seconds = await self._get_float(
                        "Ramp rate",
                        default=0.5,
                        maximum=480,
                        minimum=0.1,
                    )
            else:
                data2_seconds = data2

            data2 = seconds_to_ramp_rate(data2_seconds)
            if RAMP_RATES[data2] != data2_seconds:
                self._log_stdout(
                    f"Ramp rate rounded to {RAMP_RATES[data2]} to conform to standard values."
                )

        elif data2 is None:
            data2 = await self._get_int("Data2", default=0, values=range(0, 255))

        if data3 is None:
            if device.cat in [
                DeviceCategory.DIMMABLE_LIGHTING_CONTROL,
                DeviceCategory.SWITCHED_LIGHTING_CONTROL,
            ]:
                data3 = await self._get_int("Button", default=1, values=range(0, 255))
            else:
                data3 = await self._get_int("Data3", default=0, values=range(0, 255))

        await async_add_device_to_scene(device, scene, data1, data2, data3)

    def do_print_aldb_load_status(self, *args, **kwargs):
        """Print the All-Link databbase load status for all devices."""
        self._log_stdout("")
        self._log_stdout("Device   Status")
        self._log_stdout("-------- ---------------")
        for address in devices:
            self._log_stdout(f"{address} {str(devices[address].aldb.status)}")

    async def do_advanced(self, *args, **kwargs):
        """Enter advanced ALDB menu."""
        self._log_command("advanced")
        await self._call_next_menu(AdvancedTools, "advanced")
